// Cannon Ball Demake by SkyBerron <skyberron@gmail.com>
// This work is licensed under the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International License.
// To view a copy of the license, visit https://creativecommons.org/licenses/by-nc-nd/4.0/legalcode

const minTX = 0, maxTX = 26;
const minTY = 0, maxTY = 20;
const minX = ( ( minTX + 2 ) << 8 ), maxX = ( ( maxTX ) << 8 );
const minY = ( ( minTY + 2 ) << 8 ), maxY = ( ( maxTY ) << 8 );
const lvl = builtin("sLvl");
const pts = builtin("sPts");
const player = builtin("char14");
const wall = builtin("floor1");
const ball1 = builtin("rock");
const ball2 = builtin("sBtn");
const bgColor = ( 16 * 4 ) + 15; //129;
const playerColor = 7; //15;
const wallColor = 7;

const tRad = [ ( 4 << 5 ), ( 8 << 5 ), ( 8 << 5 ), ( 15 << 5 ) ];
const tSpr = [ ball2, ball2, ball1, ball1 ];
const tH = [ ( 40 << 5 ), ( 56 << 5 ), ( 80 << 5 ), ( 120 << 5 ) ];
const h_to_w = ( 1 << 7 );
const tS = [ 1, 2, 1, 2 ];

const cBall = 50 << 3;
var nBall;
var tBall = Array( cPlatform );
const vxBall = 1 << 5;
// [0] = x
// [1] = h
// [2] = vx
// [3] = ph
// [4] = typ

var cHarpoon = 2 << 2;
var nHarpoon;
var tHarpoon = Array( cHarpoon );
const vyHarpoon = 3 << 5;

var level = 1;
var dead;
var score;

var bx, bvx;
const by = maxY - ( 8 << 5 );
const yfloor = maxY;


reset();



function reset(){
    highscore( score );
    score = 0;

    tileshift( 0, 0 );
    fill( 250 - ( 3 << 3 ) );
    
    color( wallColor );
    for( var ty = minTX; ty <= maxTY; ty++ ) {
        tile( minTX, ty, wall );
        tile( maxTX, ty, wall );
    }
    for( var tx = minTX; tx <= maxTX; tx++ ) {
        tile( tx, minTY, wall );
        tile( tx, maxTY, wall );
    }

    bx = ( minX + maxX ) >> 1;
    nBall = level << 3;
    nHarpoon = 0;

    for( var iball = 0; iball < nBall; iball += 8 ) {
        // [0] = x
        // [1] = h
        // [2] = vx
        // [3] = ph
        // [4] = typ
        var x = random( minX, maxX );
        var v = vxBall;
        if( x > bx ) {
            v = -v;
        }
        tBall[ iball ] = x;
        tBall[ iball + 1 ] = tH[3];
        tBall[ iball + 2 ] = v;
        tBall[ iball + 3 ] = 0;
        tBall[ iball + 4 ] = 3;
    }
    
    bvx = 2 << 5;
    dead = false;
}

/*
function test_collide( x1, y1, x2, y2, ex, ey ){
    var dx = x2 - x1, dy = y2 - y1;
    return( ( dx > -ex ) && ( dx < ex ) && ( dy > -ey ) && ( dy < ey ) );
}
*/

function test_collide2( x1, y1, x2, y2, rx, ry ){
    var dx = x2 - x1, dy = y2 - y1, d = rx + ry;
    return( ( ( dx * dx ) + ( dy * dy ) ) < ( d * d ) );
}

function ball_explode( x, y, typ ) {
    if( ( typ >= 0 ) && ( nBall < ( cBall - 8 ) ) ) {
        var rad = tRad[typ];
        var h1 = yfloor - rad - y;

        // h - h1 = h * ph^2 / (k*h)^2
        // ph^2 = k^2 * h * ( h - h1 ) = k^2 * p * h^2
        // h - h1 = p * h => h = h1 / ( 1 - p )
        // ph / h = k * sqr( p )
        // if p = 1/4 => sqr( p ) = 1/2, 1 - p = 3/4
        
        var h = ( h1 << 2 ) / 3;
        var ph = ( h_to_w * h ) >> 9;
        for( var i = -1; i <= 1; i += 2 ) {
            // [0] = x
            // [1] = h
            // [2] = vx
            // [3] = ph
            // [4] = typ
            tBall[ nBall ] = x;
            tBall[ nBall + 1 ] = h;
            tBall[ nBall + 2 ] = i * vxBall;
            tBall[ nBall + 3 ] = -i * ph;
            tBall[ nBall + 4 ] = typ;
            nBall += 8;
        }
    }
}


function updateBall( iball ){
    // [0] = x
    // [1] = h
    // [2] = vx
    // [3] = ph
    // [4] = typ
    var i;
    var x = tBall[ iball ];
    var h = tBall[ iball + 1 ];
    var vx = tBall[ iball + 2 ];
    var ph = tBall[ iball + 3 ];
    var typ = tBall[ iball + 4 ];
    var rad = tRad[ typ ];
    var w = ( h_to_w * h ) >> 8;
    var h0 = tH[typ];
    var w0 = ( h_to_w * h0 ) >> 8;
    
    x += vx;
    ph += vx;
    
    if( ( ( vx < 0 ) && ( x <= ( minX + rad ) ) ) || ( ( vx > 0 ) && ( x >= ( maxX - rad ) ) ) ) {
        vx = -vx;
        ph = -ph;
    }
    
    var dy = ( ( ( h * ph ) / w ) * ph ) / w;
    if( dy > h ) {
        dy = h;
    }
    var y = yfloor - rad - h + dy;
    
    io( "SCALE", tS[ typ ] );
    sprite( ( ( x - rad ) >> 5 ), ( ( y - rad ) >> 5 ), tSpr[ typ ] );

    if( ( vx > 0 ) && ( ph > w ) ) {
        //ph -= ( w << 1 );
        ph -= w + w0;
        h = h0;
    }
    if( ( vx < 0 ) && ( ph < ( -w ) ) ) {
        //ph += ( w << 1 );
        ph += w + w0;
        h = h0;
    }

    tBall[ iball ] = x;
    tBall[ iball + 1 ] = h;
    tBall[ iball + 2 ] = vx;
    tBall[ iball + 3 ] = ph;
    
    for( var iharpoon = 0; iharpoon < nHarpoon; iharpoon += 4 ) {
        var xh = tHarpoon[ iharpoon ];
        if( ( xh > ( x - rad ) ) && ( xh < ( x + rad ) ) ) {
            var yh = yfloor - tHarpoon[ iharpoon + 1 ];
            if( ( yh < y ) || test_collide2( x, y, xh, yh, rad, ( 4 << 5 ) ) ) {
                
                for( i = iharpoon + 4; i < nHarpoon; i++ ) {
                    tHarpoon[ i - 4 ] = tHarpoon[ i ];
                }
                nHarpoon -= 4;
                
                iharpoon = nHarpoon;
                
                for( i = iball + 8; i < nBall; i++ ) {
                    tBall[ i - 8 ] = tBall[ i ];
                }
                nBall -= 8;
                score++;
                
                ball_explode( x, y, typ - 1 );
            }
        }
    }
    
    if( test_collide2( x, y, bx, by, rad, ( 6 << 5 ) ) ) {
        dead = true;
    }
    
}




function update(){
    
    for( var iball = 0; iball < nBall; iball += 8 ) {
        updateBall( iball );
    }
    io( "SCALE", 1 );
    
    for( var iharpoon = 0; iharpoon < nHarpoon; iharpoon += 4 ) {
        tHarpoon[ iharpoon + 1 ] += tHarpoon[ iharpoon + 2 ];
        var xh = tHarpoon[ iharpoon ];
        var yh = yfloor - tHarpoon[ iharpoon + 1 ];
        if( yh < minY ) {
            for( var i = iharpoon + 4; i < nHarpoon; i++ ) {
                tHarpoon[ i - 4 ] = tHarpoon[ i ]; 
            }
            nHarpoon -= 4;
        }
        var ch = "^";
        while( yh < maxY ) {
            cursor( ( xh >> 5 ) - 4, ( yh >> 5 ) - 4 );
            print( ch );
            ch = "|";
            yh += ( 8 << 5 );
        }
    }
    
    if( pressed( "C" ) ) {
        exit();
    }
    
    if( pressed( "LEFT" ) ) {
        bx -= bvx;
        var v = minX + ( 8 << 5 );
        if( bx < v ){
            bx = v;
        }
    }
    if( pressed( "RIGHT" ) ) {
        bx += bvx;
        var v = maxX - ( 8 << 5 );
        if( bx > v ) {
            bx = v;
        }
    }
    
    if( justPressed( "A" ) ) {
        if( nHarpoon < cHarpoon ) {
            tHarpoon[ nHarpoon ] = bx;
            tHarpoon[ nHarpoon + 1 ] = 0;
            tHarpoon[ nHarpoon + 2 ] = vyHarpoon;
            nHarpoon += 4;
        }
    }

    color( playerColor );
    sprite( ( ( bx ) >> 5 ) - 8, ( ( by ) >> 5 ) - 8, player );

    // hud
    //color( 47 );
    //sprite( 50, 165, lvl );
    //cursor( 70, 165 );
    //printNumber( level );
    color( 47 );
    sprite( 120, 165, pts );
    cursor( 140, 165 );
    printNumber( score );
    
    if( dead ) {
        level = 1;
        reset();
    }
    if( nBall == 0 ) {
        level++;
        reset();
    }
}

